//! Handles spawning the process
use tokio::{
    process::{
	Command,
    },
    sync::{
	mpsc,
    },
    io::{
	BufReader,
    },
    stream::StreamExt,
    prelude::*,
    task,
};
use std::{
    io,
    path::Path,
};

/// Spawn the process, and contain its standard output.
///
/// # Notes
/// Standard error is printed immediately instead.
pub async fn contained_spawn<T,U,V>(process: T, args: U, mut output_to: mpsc::Sender<String>) -> Result<(), Error>
where T: AsRef<Path>,
      U: IntoIterator<Item=V>,
      V: AsRef<std::ffi::OsStr>
{
    let mut child = match Command::new(process.as_ref())
	.args(args)
	.stdout(std::process::Stdio::piped())
	.stderr(std::process::Stdio::inherit())
	.stdin(std::process::Stdio::null())
	.spawn() {
	    Ok(chi) => chi,
	    Err(sp) => {
		return Err(Error::Spawning(sp));
	    }
	};
    let stdout = child.stdout.take().unwrap();
    let sender = tokio::spawn(async move {
	let mut lines = BufReader::new(stdout).lines();

	while let Some(Ok(line)) = lines.next().await {
	    if let Err(_) = output_to.send(line).await {
		break;
	    }
	}
    });
    task::yield_now().await;
    match child.await {
	Ok(exit) => {
	    if exit.success() {
		sender.await.expect("Child panic");
		Ok(())
	    } else {
		Err(Error::Process(exit.code()))
	    }
	},
	Err(err) => Err(Error::Transmission(err)),
    }
}

#[derive(Debug)]
pub enum Error {
    /// There was an error spawning the process
    Spawning(io::Error),
    /// Process exited with non-zero error code.
    Process(Option<i32>),
    /// Error communicating with process.
    Transmission(io::Error),
}
impl std::error::Error for Error{}
impl std::fmt::Display for Error
{
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result
    {
	match self {
	    Self::Spawning(io) => write!(f, "there was an error spawning the process: {}", io),
	    Self::Transmission(io) => write!(f, "there was an error waiting on child process: {}", io),
	    Self::Process(Some(sig)) => write!(f, "process exited with non-zero code: {}", sig),
	    Self::Process(None) => write!(f, "process teminated by signal"),
	}
    }
}
